/* cvsd: a little chroot/suid wrapper to run CVS more securely.
   Chris Black <cblack@mokey.com> http://cblack.mokey.com/
   Many ideas taken from various CVS documentation and discussions and 
   basic chroot wrapper scripts.
   This code is under the GPL, see www.gnu.org for details.
   No warranty, I take no responsibility if this allows someone to put a 
   subtle bug into your missile tracking software during development which 
   leads to Bad Things.
*/


#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>

int uid, gid;
char path[80];
char reposstr[128];
char **args;
extern int errno;
extern char **environ;

int parseline(char *line) {
  int i = 0;
  char *user, *group;
  struct passwd *passwd;
  char *chunk;
  chunk = strtok(line, " =\t");
  if(!strcasecmp(chunk, "path")) {
    strncpy(path, strtok(NULL, " =\t"), 60);
    path[strlen(path) - 1] = '\0';
  }
  if(!strcasecmp(chunk, "user")) {
    user = strtok(NULL, " =\t");
    user[strlen(user) - 1] = '\0';
    //    printf("pre getpwnam(%s)\n", user); fflush(stdout);
    passwd = getpwnam(user);
    //printf("post getpwnam\n"); fflush(stdout);
    if(passwd != NULL) {
      uid = (int) passwd->pw_uid;
      gid = (int) passwd->pw_gid;
      //printf("uid: %d, gid: %d\n", uid, gid); fflush(stdout);
    }
  }
  if(!strcasecmp(chunk, "repos")) {
    strncpy(reposstr, strtok(NULL, " =\t"), 120);
    reposstr[strlen(reposstr) - 1] = '\0';
  }
  return 1;
}

int config(void) {
  int numroots; /* number of allowable roots */
  int i;
  char rn[20];
  char buf[1024];
  static char *filename = "/etc/cvsd.conf";
  char *p;
  FILE *config;
  int **repos;

  config = fopen(filename, "r");
  
  while(fgets(buf, 1024, config)) {
    if((buf[0] != '#') && (buf[0] != '\n'))  
      parseline(buf);
  }

  if(reposstr == NULL) { snprintf(reposstr, 30, "/cvsroot"); }
  p = strchr(reposstr, ',');
  if(p == NULL) {
    /* no commas found in reposstr, only one allowed cvsroot */
    repos = (int **)malloc(sizeof(int *));
    repos[0] = (int *)malloc(strlen(reposstr) * sizeof(int *));
    snprintf((char *)repos[0], 28, "%s", reposstr);
    numroots = 1;
  } else {
    i = 2;
    p++;
    while((p = (void *)strchr(p, ',')) != NULL) {
      i++; p++;
    }
    numroots = i;
    i = 0;
    repos = (int **)malloc(numroots * sizeof(int *));
    p = strtok(reposstr, ",");
    strncpy(rn, p, 19);
    repos[i] = (int *)malloc(strlen(rn) * sizeof(int));
    strncpy((char *)repos[i], rn, 20);
    for(i = 1; i < numroots; i++) {
      p = strtok(NULL, ",");
      strncpy(rn, p, 19);
      repos[i] = (int *)malloc(strlen(rn) * sizeof(int));
      strncpy((char *)repos[i], rn, 20);
    }
  }

  /* build argument list from repos array */
  args = (char **)malloc((numroots + 2) * sizeof(char));
  args[0] = "cvs";
  for(i = 1; i < numroots + 1; i++) {
    args[i] = (char *)
      malloc((strlen((char *)repos[i-1]) + 15) * sizeof(char));
    snprintf((char *)args[i], 40, "--allow-root=%s", (char *)repos[i-1]);
    free(repos[i-1]);
  }
  args[i] = "pserver";
  args[i + 1] = NULL;

  free(repos);
  
  return 1;
}

int main(int argc, char *argv[], char *env[]) {
  int i;
  char *buf, *p, *cvsroot;

  openlog("cvsd", 0, LOG_DAEMON);
  syslog(LOG_INFO, "Connection made.\n");

  config();

  //printf("path = %s\n", path);
  //printf("uid = %d, gid = %d\n", uid, gid);

  /*
  for(i = 0; args[i] != NULL; i++) {
    printf("args[%d]: %s\n", i, args[i]);
  }
  */

  /* peek at auth block */
  buf = (char *)malloc(1024 * sizeof(char));
  recv(STDIN_FILENO, buf, 1024, MSG_PEEK);
  //syslog(LOG_INFO, "peek buf: %s\n", buf);

  /* get cvsroot */
  p = strchr(buf, '/');
  cvsroot = strtok(p, " \r\n");
  syslog(LOG_INFO, "cvsroot: \"%s\"\n", cvsroot);
  free(buf);

  //recv(STDIN_FILENO, buf, 1024, 0);
  //syslog(LOG_INFO, "after peek buf: %s\n", buf);


  if(chdir(path)) { 
    syslog(LOG_ERR, "Could not chdir to %s\n", path); exit(1); }
  if(chroot(path)) { 
    syslog(LOG_ERR, "Could not chroot to %s\n", path); exit(2); }
  if(setgid(gid)) { 
    syslog(LOG_ERR, "Could not setgid to %d\n", gid); exit(3); }
  if(setuid(uid)) { 
    syslog(LOG_ERR, "Could not setuid to %d\n", uid); exit(4); }
  
  execvp("cvs", args);

  syslog(LOG_ERR, "Error executing cvs");

  exit(5);
}
